Package: ManifoldOptim
Type: Package
Title: An R Interface to the 'ROPTLIB' Library for Riemannian Manifold
        Optimization
Version: 1.0.2
Authors@R: c(person("Kofi P. Adragni", role = c("aut","cph")),
    person("Sean R. Martin", role = c("aut","cre","cph"),
        email = "sean.martin@jhuapl.edu"),
    person("Andrew M. Raim", role = c("aut","cph"),
        email = "andrew.raim@gmail.com"),
    person("Wen Huang", role = c("aut","cph")))
Description: An R interface to version 0.3 of the 'ROPTLIB' optimization library
    (see <https://www.math.fsu.edu/~whuang2/> for more information). Optimize
    real-valued functions over manifolds such as Stiefel, Grassmann, and
    symmetric positive definite matrices. For details see Martin et al. (2020)
    <doi:10.18637/jss.v093.i01>. Note that the optional 'ldr' package used in
    some of this package's examples can be obtained from either the article
    <doi:10.18637/jss.v061.i03> or from the 'ldr' package
    <https://cran.r-project.org/package=ldr>.
License: GPL (>= 2)
URL: https://bitbucket.org/umbc_sdr/manifoldoptim
Depends: Rcpp (>= 0.12.0)
Suggests: RcppArmadillo
LinkingTo: Rcpp, RcppArmadillo
RcppModules: ManifoldOptim_module
RoxygenNote: 7.3.2
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2026-07-12 14:19:15 UTC; test
Author: Kofi P. Adragni [aut, cph],
  Sean R. Martin [aut, cre, cph],
  Andrew M. Raim [aut, cph],
  Wen Huang [aut, cph]
Maintainer: Sean R. Martin <sean.martin@jhuapl.edu>
Repository: CRAN
Date/Publication: 2026-07-12 18:20:15 UTC
Built: R 4.4.3; x86_64-apple-darwin20; 2026-07-12 21:33:58 UTC; unix
Archs: ManifoldOptim.so.dSYM
