# Generated by roxygen2: do not edit by hand

S3method(print,ManifoldOptim)
export(Trace)
export(get.deriv.params)
export(get.euclidean.defn)
export(get.grassmann.defn)
export(get.lowrank.defn)
export(get.manifold.params)
export(get.orthgroup.defn)
export(get.product.defn)
export(get.solver.params)
export(get.spd.defn)
export(get.sphere.defn)
export(get.stiefel.defn)
export(manifold.optim)
export(moptim)
export(orthonorm)
importFrom(Rcpp,evalCpp)
useDynLib(ManifoldOptim, .registration = TRUE)
